/*

	GDI32 video backend, uses bitmap and StretchDIBits for rendering
 
*/

#include <windows.h>
#include <string.h>
#include "m_gfx.h"
#include "m_core.h"

// forward definition of palette in BGRA format - RGBQUAD
extern unsigned char GamePalette[1024];

typedef struct tagBITMAPINFO2 {
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD bmiColors[256];
} BITMAPINFO2;

BITMAPINFO2 bmp;
HINSTANCE hInst;
HWND hWnd;
MSG msg;
char AppTitle[] = "The Last Mission Win32 GDI Remake";

unsigned char Keys[128];
unsigned char ScreenBuffer[320*200];
unsigned char ScreenBigBuffer[640*400];
int scale2x = 2; // 1 - no scale 320x200; 2 - upscale to 640x400

int WndCreate();
void LM_GFX_Init();

//================================================================
void LM_ResetKeys()
{
	memset(&Keys[0], 0, 128);
}

int LM_AnyKey()
{
    for(int i = 0; i < 127; i++)
	{
        if(Keys[i] == 1) return 1;
    }
    return 0;
}

int LM_Timer()
{
    return GetTickCount();
}

void LM_Sleep(int sleep_time)
{
    Sleep(sleep_time);
}

// pScreenBuffer - a pointer to a pointer to a buffer 
int LM_Init(unsigned char **pScreenBuffer)
{
    if(WndCreate() == 0) return 0;
    LM_GFX_Init();
	*pScreenBuffer = &ScreenBuffer[0];
	return 1;
}

void LM_Deinit()
{
    DestroyWindow(hWnd);
}

char LM_PollEvents()
{
    if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) != 0)
	{
      TranslateMessage(&msg);
      DispatchMessage(&msg);
      if(msg.message == WM_QUIT) return 1;
    }

    return 0;
}

int _toggle = 1;

// hint: CALLBACK is absolutely necessary for watcom, mingw compiles fine without it
LRESULT CALLBACK WndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{    
	switch(uMsg)
	{
        case WM_CREATE:
			break;
        case WM_DESTROY:
            PostQuitMessage(0);
			break;
        case WM_CLOSE:
            LM_Deinit();
            ExitProcess(0);
			break;
        case WM_KEYUP:
			Keys[(lParam >> 16) & 0x7f] = 0;
			break;
        case WM_KEYDOWN:
			Keys[(lParam >> 16) & 0x7f] = 1;
			// toggle scaler system/manual
			if(Keys[SC_S] == 1) {
				Keys[SC_S] = 0; 
				_toggle ^= 1; 
				LM_GFX_SetScale(_toggle + 1);			
			}
    	}
		
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

int WndCreate()
{
	WNDCLASS wc;
	RECT rect;

	// hint: mingw compiles & exe works without the following string, but watcom not
	hInst = GetModuleHandle(0);    
	if(hInst == 0) return 0;
	
	wc.style          = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc    = (WNDPROC)WndProc;
	wc.cbClsExtra     = 0;
	wc.cbWndExtra     = 0;
	wc.hInstance      = hInst;
	wc.hIcon          = NULL;
	wc.hCursor        = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground  = (HBRUSH)COLOR_WINDOW;
	wc.lpszMenuName   = NULL;
	wc.lpszClassName  = AppTitle;
   	
	RegisterClass(&wc);

    rect.top = 0;
    rect.bottom = 400;
    rect.left = 0;
    rect.right = 640;
   	
    AdjustWindowRect(&rect, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, FALSE);
		
	hWnd = CreateWindow( 	AppTitle,
							AppTitle,
							WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
							CW_USEDEFAULT,
							CW_USEDEFAULT,
							rect.right - rect.left,
							rect.bottom - rect.top,
							NULL,
							NULL,
							hInst,
							NULL);	
		
	if(hWnd != 0)
	{
        ShowWindow(hWnd, SW_SHOW);
        UpdateWindow(hWnd);
		return 1;
    } 
		
	return 0;
}

void LM_GFX_Init()
{

    if(GetDC(hWnd) == 0) return;
    ReleaseDC(hWnd, 0);
    
    memset(&bmp, 0, sizeof(bmp));
        
    bmp.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmp.bmiHeader.biWidth = 320 * scale2x;
    bmp.bmiHeader.biHeight = -200 * scale2x;
    bmp.bmiHeader.biPlanes = 1;
    bmp.bmiHeader.biBitCount = 8;
    bmp.bmiHeader.biCompression = BI_RGB;
    
    memcpy(&bmp.bmiColors[0], &GamePalette[0], 1024);

}



void LM_GFX_Deinit()
{
}

void LM_GFX_Flip(unsigned char *p)
{
    RECT client;
	HDC hDC;
    
    hDC = GetDC(hWnd);
	
    if(hDC != 0)
	{
        // double 320x200 to 640x200
        if(scale2x == 2)
		{
            for(int y = 0; y <= 199; y++)
			{
                for(int x = 0; x <= 319; x++)
				{
                    unsigned char tmp = *(unsigned char *)(p + y*320+x);
    
                    ScreenBigBuffer[(y*2+1)*640+x*2] = tmp;
                    ScreenBigBuffer[(y*2+1)*640+x*2+1] = tmp;
                    // dirty scanlines :) not needed in 256 colors
                    //if((tmp <= 8) && (tmp != 0)) tmp += 8;
                    ScreenBigBuffer[y*2*640+x*2] = tmp;
                    ScreenBigBuffer[y*2*640+x*2+1] = tmp;
                }
            }

        }

        GetClientRect(hWnd, &client);
        client.right = 320 * scale2x;
        client.bottom = 200 * scale2x;
		
        StretchDIBits(	hDC, 
						0,								// Destination top left hand corner X Position
						0,								// Destination top left hand corner Y Position
						client.right,					// Destinations Width
						client.bottom,					// Desitnations height
						0,								// Source top left hand corner's X Position
						0,								// Source top left hand corner's Y Position
						320 * scale2x,					// Sources width
						200 * scale2x,					// Sources height
						(scale2x == 1 ? p : &ScreenBigBuffer[0]),   // Source's data
						(BITMAPINFO *)&bmp,				// Bitmap Info
						DIB_RGB_COLORS,					// operations
						SRCCOPY);

                 
        ReleaseDC(hWnd, hDC);
    }
}

void LM_GFX_WaitVSync()
{
    // not needed in WinGDI
}


void LM_GFX_SetScale(int param)
{
    RECT rect;
	HDC hDC;
	
	param = ((param - 1) & 1) + 1;
    scale2x = param;
    bmp.bmiHeader.biWidth = 320 * param;
    bmp.bmiHeader.biHeight = -200 * param;
	
	hDC = GetDC(hWnd);
	GetClientRect(hWnd, &rect);
	rect.top = 0;
    rect.bottom = 200 * param;
    rect.left = 0;
    rect.right = 320 * param;
   	
    AdjustWindowRect(&rect, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, FALSE);	
	
	// Shitty function, had to bruteforce GetSystemMetrics
	SetWindowPos(	hWnd, 0, 0, 0, 
					rect.right + GetSystemMetrics(SM_CXFRAME)/2, // i don't understand why /2 but it works thus i don't care
					rect.bottom + GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYFRAME)/2, 
					SWP_NOMOVE);
	ReleaseDC(hWnd, hDC);
	
}

// palette must be in rgba format
/*void LM_GFX_SetPalette(unsigned char *p)
{
}*/

// palette in bgra format
unsigned char GamePalette[1024] =
{
    0x00, 0x00, 0x00, 0x00, 0xAA, 0x00, 0x00, 0x00, 0x00, 0xAA, 0x00, 0x00, 0xAA, 0xAA, 0x00, 0x00, 
    0x00, 0x00, 0xAA, 0x00, 0xAA, 0x00, 0xAA, 0x00, 0x00, 0x55, 0xAA, 0x00, 0xAA, 0xAA, 0xAA, 0x00, 
    0x55, 0x55, 0x55, 0x00, 0xFF, 0x55, 0x55, 0x00, 0x55, 0xFF, 0x55, 0x00, 0xFF, 0xFF, 0x55, 0x00, 
    0x55, 0x55, 0xFF, 0x00, 0xFF, 0x55, 0xFF, 0x00, 0x55, 0xFF, 0xFF, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x14, 0x14, 0x14, 0x00, 0x20, 0x20, 0x20, 0x00, 0x2C, 0x2C, 0x2C, 0x00, 
    0x38, 0x38, 0x38, 0x00, 0x44, 0x44, 0x44, 0x00, 0x50, 0x50, 0x50, 0x00, 0x61, 0x61, 0x61, 0x00, 
    0x71, 0x71, 0x71, 0x00, 0x81, 0x81, 0x81, 0x00, 0x91, 0x91, 0x91, 0x00, 0xA1, 0xA1, 0xA1, 0x00, 
    0xB6, 0xB6, 0xB6, 0x00, 0xCA, 0xCA, 0xCA, 0x00, 0xE2, 0xE2, 0xE2, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 
    0xFF, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x40, 0x00, 0xFF, 0x00, 0x7D, 0x00, 0xFF, 0x00, 0xBE, 0x00, 
    0xFF, 0x00, 0xFF, 0x00, 0xBE, 0x00, 0xFF, 0x00, 0x7D, 0x00, 0xFF, 0x00, 0x40, 0x00, 0xFF, 0x00, 
    0x00, 0x00, 0xFF, 0x00, 0x00, 0x40, 0xFF, 0x00, 0x00, 0x7D, 0xFF, 0x00, 0x00, 0xBE, 0xFF, 0x00, 
    0x00, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xBE, 0x00, 0x00, 0xFF, 0x7D, 0x00, 0x00, 0xFF, 0x40, 0x00, 
    0x00, 0xFF, 0x00, 0x00, 0x40, 0xFF, 0x00, 0x00, 0x7D, 0xFF, 0x00, 0x00, 0xBE, 0xFF, 0x00, 0x00, 
    0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xBE, 0x00, 0x00, 0xFF, 0x7D, 0x00, 0x00, 0xFF, 0x40, 0x00, 0x00, 
    0xFF, 0x7D, 0x7D, 0x00, 0xFF, 0x7D, 0x9D, 0x00, 0xFF, 0x7D, 0xBE, 0x00, 0xFF, 0x7D, 0xDE, 0x00, 
    0xFF, 0x7D, 0xFF, 0x00, 0xDE, 0x7D, 0xFF, 0x00, 0xBE, 0x7D, 0xFF, 0x00, 0x9D, 0x7D, 0xFF, 0x00, 
    0x7D, 0x7D, 0xFF, 0x00, 0x7D, 0x9D, 0xFF, 0x00, 0x7D, 0xBE, 0xFF, 0x00, 0x7D, 0xDE, 0xFF, 0x00, 
    0x7D, 0xFF, 0xFF, 0x00, 0x7D, 0xFF, 0xDE, 0x00, 0x7D, 0xFF, 0xBE, 0x00, 0x7D, 0xFF, 0x9D, 0x00, 
    0x7D, 0xFF, 0x7D, 0x00, 0x9D, 0xFF, 0x7D, 0x00, 0xBE, 0xFF, 0x7D, 0x00, 0xDE, 0xFF, 0x7D, 0x00, 
    0xFF, 0xFF, 0x7D, 0x00, 0xFF, 0xDE, 0x7D, 0x00, 0xFF, 0xBE, 0x7D, 0x00, 0xFF, 0x9D, 0x7D, 0x00, 
    0xFF, 0xB6, 0xB6, 0x00, 0xFF, 0xB6, 0xC6, 0x00, 0xFF, 0xB6, 0xDA, 0x00, 0xFF, 0xB6, 0xEA, 0x00, 
    0xFF, 0xB6, 0xFF, 0x00, 0xEA, 0xB6, 0xFF, 0x00, 0xDA, 0xB6, 0xFF, 0x00, 0xC6, 0xB6, 0xFF, 0x00, 
    0xB6, 0xB6, 0xFF, 0x00, 0xB6, 0xC6, 0xFF, 0x00, 0xB6, 0xDA, 0xFF, 0x00, 0xB6, 0xEA, 0xFF, 0x00, 
    0xB6, 0xFF, 0xFF, 0x00, 0xB6, 0xFF, 0xEA, 0x00, 0xB6, 0xFF, 0xDA, 0x00, 0xB6, 0xFF, 0xC6, 0x00, 
    0xB6, 0xFF, 0xB6, 0x00, 0xC6, 0xFF, 0xB6, 0x00, 0xDA, 0xFF, 0xB6, 0x00, 0xEA, 0xFF, 0xB6, 0x00, 
    0xFF, 0xFF, 0xB6, 0x00, 0xFF, 0xEA, 0xB6, 0x00, 0xFF, 0xDA, 0xB6, 0x00, 0xFF, 0xC6, 0xB6, 0x00, 
    0x71, 0x00, 0x00, 0x00, 0x71, 0x00, 0x1C, 0x00, 0x71, 0x00, 0x38, 0x00, 0x71, 0x00, 0x55, 0x00, 
    0x71, 0x00, 0x71, 0x00, 0x55, 0x00, 0x71, 0x00, 0x38, 0x00, 0x71, 0x00, 0x1C, 0x00, 0x71, 0x00, 
    0x00, 0x00, 0x71, 0x00, 0x00, 0x1C, 0x71, 0x00, 0x00, 0x38, 0x71, 0x00, 0x00, 0x55, 0x71, 0x00, 
    0x00, 0x71, 0x71, 0x00, 0x00, 0x71, 0x55, 0x00, 0x00, 0x71, 0x38, 0x00, 0x00, 0x71, 0x1C, 0x00, 
    0x00, 0x71, 0x00, 0x00, 0x1C, 0x71, 0x00, 0x00, 0x38, 0x71, 0x00, 0x00, 0x55, 0x71, 0x00, 0x00, 
    0x71, 0x71, 0x00, 0x00, 0x71, 0x55, 0x00, 0x00, 0x71, 0x38, 0x00, 0x00, 0x71, 0x1C, 0x00, 0x00, 
    0x71, 0x38, 0x38, 0x00, 0x71, 0x38, 0x44, 0x00, 0x71, 0x38, 0x55, 0x00, 0x71, 0x38, 0x61, 0x00, 
    0x71, 0x38, 0x71, 0x00, 0x61, 0x38, 0x71, 0x00, 0x55, 0x38, 0x71, 0x00, 0x44, 0x38, 0x71, 0x00, 
    0x38, 0x38, 0x71, 0x00, 0x38, 0x44, 0x71, 0x00, 0x38, 0x55, 0x71, 0x00, 0x38, 0x61, 0x71, 0x00, 
    0x38, 0x71, 0x71, 0x00, 0x38, 0x71, 0x61, 0x00, 0x38, 0x71, 0x55, 0x00, 0x38, 0x71, 0x44, 0x00, 
    0x38, 0x71, 0x38, 0x00, 0x44, 0x71, 0x38, 0x00, 0x55, 0x71, 0x38, 0x00, 0x61, 0x71, 0x38, 0x00, 
    0x71, 0x71, 0x38, 0x00, 0x71, 0x61, 0x38, 0x00, 0x71, 0x55, 0x38, 0x00, 0x71, 0x44, 0x38, 0x00, 
    0x71, 0x50, 0x50, 0x00, 0x71, 0x50, 0x59, 0x00, 0x71, 0x50, 0x61, 0x00, 0x71, 0x50, 0x69, 0x00, 
    0x71, 0x50, 0x71, 0x00, 0x69, 0x50, 0x71, 0x00, 0x61, 0x50, 0x71, 0x00, 0x59, 0x50, 0x71, 0x00, 
    0x50, 0x50, 0x71, 0x00, 0x50, 0x59, 0x71, 0x00, 0x50, 0x61, 0x71, 0x00, 0x50, 0x69, 0x71, 0x00, 
    0x50, 0x71, 0x71, 0x00, 0x50, 0x71, 0x69, 0x00, 0x50, 0x71, 0x61, 0x00, 0x50, 0x71, 0x59, 0x00, 
    0x50, 0x71, 0x50, 0x00, 0x59, 0x71, 0x50, 0x00, 0x61, 0x71, 0x50, 0x00, 0x69, 0x71, 0x50, 0x00, 
    0x71, 0x71, 0x50, 0x00, 0x71, 0x69, 0x50, 0x00, 0x71, 0x61, 0x50, 0x00, 0x71, 0x59, 0x50, 0x00, 
    0x40, 0x00, 0x00, 0x00, 0x40, 0x00, 0x10, 0x00, 0x40, 0x00, 0x20, 0x00, 0x40, 0x00, 0x30, 0x00, 
    0x40, 0x00, 0x40, 0x00, 0x30, 0x00, 0x40, 0x00, 0x20, 0x00, 0x40, 0x00, 0x10, 0x00, 0x40, 0x00, 
    0x00, 0x00, 0x40, 0x00, 0x00, 0x10, 0x40, 0x00, 0x00, 0x20, 0x40, 0x00, 0x00, 0x30, 0x40, 0x00, 
    0x00, 0x40, 0x40, 0x00, 0x00, 0x40, 0x30, 0x00, 0x00, 0x40, 0x20, 0x00, 0x00, 0x40, 0x10, 0x00, 
    0x00, 0x40, 0x00, 0x00, 0x10, 0x40, 0x00, 0x00, 0x20, 0x40, 0x00, 0x00, 0x30, 0x40, 0x00, 0x00, 
    0x40, 0x40, 0x00, 0x00, 0x40, 0x30, 0x00, 0x00, 0x40, 0x20, 0x00, 0x00, 0x40, 0x10, 0x00, 0x00, 
    0x40, 0x20, 0x20, 0x00, 0x40, 0x20, 0x28, 0x00, 0x40, 0x20, 0x30, 0x00, 0x40, 0x20, 0x38, 0x00, 
    0x40, 0x20, 0x40, 0x00, 0x38, 0x20, 0x40, 0x00, 0x30, 0x20, 0x40, 0x00, 0x28, 0x20, 0x40, 0x00, 
    0x20, 0x20, 0x40, 0x00, 0x20, 0x28, 0x40, 0x00, 0x20, 0x30, 0x40, 0x00, 0x20, 0x38, 0x40, 0x00, 
    0x20, 0x40, 0x40, 0x00, 0x20, 0x40, 0x38, 0x00, 0x20, 0x40, 0x30, 0x00, 0x20, 0x40, 0x28, 0x00, 
    0x20, 0x40, 0x20, 0x00, 0x28, 0x40, 0x20, 0x00, 0x30, 0x40, 0x20, 0x00, 0x38, 0x40, 0x20, 0x00, 
    0x40, 0x40, 0x20, 0x00, 0x40, 0x38, 0x20, 0x00, 0x40, 0x30, 0x20, 0x00, 0x40, 0x28, 0x20, 0x00, 
    0x40, 0x2C, 0x2C, 0x00, 0x40, 0x2C, 0x30, 0x00, 0x40, 0x2C, 0x34, 0x00, 0x40, 0x2C, 0x3C, 0x00, 
    0x40, 0x2C, 0x40, 0x00, 0x3C, 0x2C, 0x40, 0x00, 0x34, 0x2C, 0x40, 0x00, 0x30, 0x2C, 0x40, 0x00, 
    0x2C, 0x2C, 0x40, 0x00, 0x2C, 0x30, 0x40, 0x00, 0x2C, 0x34, 0x40, 0x00, 0x2C, 0x3C, 0x40, 0x00, 
    0x2C, 0x40, 0x40, 0x00, 0x2C, 0x40, 0x3C, 0x00, 0x2C, 0x40, 0x34, 0x00, 0x2C, 0x40, 0x30, 0x00, 
    0x2C, 0x40, 0x2C, 0x00, 0x30, 0x40, 0x2C, 0x00, 0x34, 0x40, 0x2C, 0x00, 0x3C, 0x40, 0x2C, 0x00, 
    0x40, 0x40, 0x2C, 0x00, 0x40, 0x3C, 0x2C, 0x00, 0x40, 0x34, 0x2C, 0x00, 0x40, 0x30, 0x2C, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
};











